unit AudioSource;

interface

uses
  Windows,CosineTable,Classes,WaveOut,MMSystem;

type
  TAudioSource=class
  protected
    function GetExportLength:TFloat;virtual;abstract;
    procedure Reset;virtual;abstract;
  public
    procedure Render(Buffer:Pointer;Length,Frequency:Cardinal;Stereo,Bits16:Boolean);virtual;abstract;

    procedure ExportToStream(AStream:TStream;Frequency:Cardinal;Stereo,Bits16:Boolean);

    property ExportLength:TFloat read GetExportLength;
  end;

  TDiscreteAudioSource=class(TAudioSource)
  private
    FLastT:TFloat;
  protected
    procedure Sample1(T:TFloat;var M:TFloat);virtual;abstract;
    procedure Sample2(T:TFloat;var L,R:TFloat);virtual;abstract;
  public
    procedure Render(Buffer:Pointer;Length,Frequency:Cardinal;Stereo,Bits16:Boolean);override;
  end;

  TMonoDiscreteAudioSource=class(TDiscreteAudioSource)
  private
    FVolume:TFloat;
  protected
    procedure Sample(T:TFloat;var M:TFloat);virtual;abstract;
  public
    constructor Create;

    procedure Sample1(T:TFloat;var M:TFloat);override;
    procedure Sample2(T:TFloat;var L,R:TFloat);override;

    property Volume:TFloat read FVolume write FVolume;
  end;

  TStereoDiscreteAudioSource=class(TDiscreteAudioSource)
  private
    FLVolume,FRVolume:TFloat;
  protected
    procedure Sample(T:TFloat;var L,R:TFloat);virtual;abstract;
  public
    constructor Create;

    procedure Sample1(T:TFloat;var M:TFloat);override;
    procedure Sample2(T:TFloat;var L,R:TFloat);override;

    property LVolume:TFloat read FLVolume write FLVolume;
    property RVolume:TFloat read FRVolume write FRVolume;
  end;

  TFloats=array[0..$FFFFFF] of TFloat;
  PFloats=^TFloats;

  TBufferAudioSource=class(TAudioSource)
  private
    FSamples:Cardinal;
  protected
    procedure AllocateSamples;virtual;abstract;
    procedure UnallocateSamples;virtual;abstract;

    procedure Resize(Samples:Cardinal);

    property Samples:Cardinal read FSamples;
  public
    destructor Destroy;override;
  end;

  TMonoBufferAudioSource=class(TBufferAudioSource)
  private
    FM:PFloats;
    FVolume:TFloat;
  protected
    procedure AllocateSamples;override;
    procedure UnallocateSamples;override;

    procedure Fill(M:PFloats;Samples,Frequency:Cardinal);virtual;abstract;
  public
    constructor Create;

    procedure Render(Buffer:Pointer;Length,Frequency:Cardinal;Stereo,Bits16:Boolean);override;

    property Volume:TFloat read FVolume write FVolume;
  end;

  TStereoBufferAudioSource=class(TBufferAudioSource)
  private
    FL,FR:PFloats;
    FLVolume,FRVolume:TFloat;
  protected
    procedure AllocateSamples;override;
    procedure UnallocateSamples;override;

    procedure Fill(L,R:PFloats;Samples,Frequency:Cardinal);virtual;abstract;
  public
    constructor Create;

    procedure Render(Buffer:Pointer;Length,Frequency:Cardinal;Stereo,Bits16:Boolean);override;

    property LVolume:TFloat read FLVolume write FLVolume;
    property RVolume:TFloat read FRVolume write FRVolume;
  end;

procedure Convert8(const Src:TFloat;var Dest:Byte);register;
procedure Convert16(const Src:TFloat;var Dest:Smallint);register;

implementation

procedure Convert8(const Src:TFloat;var Dest:Byte);register;
begin
  if Src>1 then
    Dest:=$FF
  else begin
    if Src<-1 then
      Dest:=0
    else
      Dest:=Round(Src*127.5+127.5);
  end;
end;

procedure Convert16(const Src:TFloat;var Dest:Smallint);register;
begin
  if Src>1 then
    Dest:=$7FFF
  else begin
    if Src<-1 then
      Dest:=-$7FFF
    else
      Dest:=Round(Src*$7FFF);
  end;
end;

{ TAudioSource }

procedure TAudioSource.ExportToStream(AStream: TStream;
  Frequency: Cardinal; Stereo, Bits16: Boolean);
const
  sRIFF='RIFF';
  sWAVE='WAVE';
  sFMT_='fmt ';
  sDATA='data';
  BuffSize=$10000;
var
  n,o:Cardinal;
  p:array[0..BuffSize-1] of Byte;

  procedure WriteWORD(w:WORD);
  begin
    AStream.WriteBuffer(w,SizeOf(w));
  end;

  procedure WriteDWORD(d:DWORD);
  begin
    AStream.WriteBuffer(d,SizeOf(d));
  end;

  procedure WriteString(s:string);
  begin
    AStream.WriteBuffer(s[1],Length(s));
  end;

begin
  Reset;
  o:=1;
  if Stereo then
    o:=2*o;
  if Bits16 then
    o:=2*o;
  n:=Round(o*ExportLength*Frequency);
  WriteString(sRIFF);
  WriteDWORD(Length(sWAVE)+Length(sFMT_)+SizeOf(DWORD)+2*SizeOf(Word)+2*SizeOf(DWORD)+2*SizeOf(Word)+Length(sDATA)+SizeOf(DWORD)+n);
  WriteString(sWAVE);
  WriteString(sFMT_);
  WriteDWORD(2*SizeOf(Word)+2*SizeOf(DWORD)+2*SizeOf(Word));
  WriteWORD(WAVE_FORMAT_PCM);
  if Stereo then
    WriteWORD(2)
  else
    WriteWORD(1);
  WriteDWORD(Frequency);
  WriteDWORD(Frequency*o);
  WriteWord(o);
  if bits16 then
    WriteWord(16)
  else
    WriteWord(8);
  WriteString(sDATA);
  WriteDWORD(n);
  while n>0 do begin
    if n<BuffSize then
      o:=n
    else
      o:=BuffSize;
    Render(@p,o,Frequency,Stereo,Bits16);
    AStream.WriteBuffer(p,o);
    Dec(n,o);
  end;
end;

{ TDiscreteAudioSource }

procedure TDiscreteAudioSource.Render(Buffer: Pointer; Length,
  Frequency: Cardinal; Stereo, Bits16: Boolean);
var
  a:Integer;
  T,L,R,M,P:TFloat;
  K:PChar;
begin
  P:=1/Frequency;
  K:=Buffer;
  if Stereo then begin
    Length:=Length div 2;
    if Bits16 then begin
      Length:=Length div 2;
      for a:=0 to Length-1 do begin
        T:=FLastT+a*P;
        Sample2(T,L,R);
        Convert16(L,PSmallint(K)^);
        Convert16(R,PSmallint(K+2)^);
        Inc(K,4);
      end;
    end else begin
      for a:=0 to Length-1 do begin
        T:=FLastT+a*P;
        Sample2(T,L,R);
        Convert8(L,PByte(K)^);
        Convert8(R,PByte(K+1)^);
        Inc(K,2);
      end;
    end;
  end else begin
    if Bits16 then begin
      Length:=Length div 2;
      for a:=0 to Length-1 do begin
        T:=FLastT+a*P;
        Sample1(T,M);
        Convert16(M,PSmallint(K)^);
        Inc(K,2);
      end;
    end else begin
      for a:=0 to Length-1 do begin
        T:=FLastT+a*P;
        Sample1(T,M);
        Convert8(M,PByte(K)^);
        Inc(K);
      end;
    end;
  end;
  FLastT:=FLastT+Length*P;
end;

{ TMonoDiscreteAudioSource }

constructor TMonoDiscreteAudioSource.Create;
begin
  FVolume:=0.3;
end;

procedure TMonoDiscreteAudioSource.Sample1(T: TFloat; var M: TFloat);
begin
  Sample(T,M);
  M:=M*FVolume;
end;

procedure TMonoDiscreteAudioSource.Sample2(T: TFloat; var L,
  R: TFloat);
begin
  Sample(T,L);
  L:=L*FVolume;
  R:=L;
end;

{ TStereoDiscreteAudioSource }

constructor TStereoDiscreteAudioSource.Create;
begin
  FLVolume:=0.5;
  FRVolume:=0.5;
end;

procedure TStereoDiscreteAudioSource.Sample1(T: TFloat; var M: TFloat);
var
  L,R:TFloat;
begin
  Sample(T,L,R);
  M:=0.5*(FLVolume*L+FRVolume*R);
end;

procedure TStereoDiscreteAudioSource.Sample2(T: TFloat; var L,
  R: TFloat);
begin
  Sample(T,L,R);
  L:=L*FLVolume;
  R:=R*FLVolume;
end;

{ TBufferAudioSource }

destructor TBufferAudioSource.Destroy;
begin
  if FSamples>0 then
    UnallocateSamples;
  inherited;
end;

procedure TBufferAudioSource.Resize(Samples: Cardinal);
begin
  if Samples<>FSamples then begin
    if FSamples>0 then
      UnallocateSamples;
    FSamples:=Samples;
    AllocateSamples;
  end;
end;

{ TMonoBufferAudioSource }

procedure TMonoBufferAudioSource.AllocateSamples;
begin
  GetMem(FM,Samples*Sizeof(TFloat));
end;

constructor TMonoBufferAudioSource.Create;
begin
  FVolume:=0.5;
end;

procedure TMonoBufferAudioSource.Render(Buffer: Pointer; Length,
  Frequency: Cardinal; Stereo, Bits16: Boolean);
var
  a:Integer;
  K:PChar;
  s:Smallint;
  b:Byte;
begin
  K:=Buffer;
  if Stereo then begin
    Length:=Length div 2;
    if Bits16 then begin
      Length:=Length div 2;
      Resize(Length);
      Fill(FM,Length,Frequency);
      for a:=0 to Length-1 do begin
        Convert16(FM[a]*FVolume,s);
        PSmallint(K)^:=s;
        PSmallint(K+2)^:=s;
        Inc(K,4);
      end;
    end else begin
      Resize(Length);
      Fill(FM,Length,Frequency);
      for a:=0 to Length-1 do begin
        Convert8(FM[a]*FVolume,b);
        PByte(K)^:=b;
        PByte(K+1)^:=b;
        Inc(K,2);
      end;
    end;
  end else begin
    if Bits16 then begin
      Length:=Length div 2;
      Resize(Length);
      Fill(FM,Length,Frequency);
      for a:=0 to Length-1 do begin
        Convert16(FM[a]*FVolume,PSmallint(K)^);
        Inc(K,2);
      end;
    end else begin
      Resize(Length);
      Fill(FM,Length,Frequency);
      for a:=0 to Length-1 do begin
        Convert8(FM[a]*FVolume,PByte(K)^);
        Inc(K);
      end;
    end;
  end;
end;

procedure TMonoBufferAudioSource.UnallocateSamples;
begin
  FreeMem(FM);
end;

{ TStereoBufferAudioSource }

procedure TStereoBufferAudioSource.AllocateSamples;
begin
  GetMem(FL,Samples*SizeOf(TFloat));
  GetMem(FR,Samples*SizeOf(TFloat));
end;

constructor TStereoBufferAudioSource.Create;
begin
  FLVolume:=0.5;
  FRVolume:=0.5;
end;

procedure TStereoBufferAudioSource.Render(Buffer: Pointer; Length,
  Frequency: Cardinal; Stereo, Bits16: Boolean);
var
  a:Integer;
  K:PChar;
begin
  K:=Buffer;
  if Stereo then begin
    Length:=Length div 2;
    if Bits16 then begin
      Length:=Length div 2;
      Resize(Length);
      Fill(FL,FR,Length,Frequency);
      for a:=0 to Length-1 do begin
        Convert16(FL[a]*FLVolume,PSmallint(K)^);
        Convert16(FR[a]*FRVolume,PSmallint(K+2)^);
        Inc(K,4);
      end;
    end else begin
      Resize(Length);
      Fill(FL,FR,Length,Frequency);
      for a:=0 to Length-1 do begin
        Convert8(FL[a]*FLVolume,PByte(K)^);
        Convert8(FR[a]*FRVolume,PByte(K+1)^);
        Inc(K,2);
      end;
    end;
  end else begin
    if Bits16 then begin
      Length:=Length div 2;
      Resize(Length);
      Fill(FL,FR,Length,Frequency);
      for a:=0 to Length-1 do begin
        Convert16(0.5*(FL[a]*FLVolume+FR[a]*FRVolume),PSmallint(K)^);
        Inc(K,2);
      end;
    end else begin
      Resize(Length);
      Fill(FL,FR,Length,Frequency);
      for a:=0 to Length-1 do begin
        Convert8(0.5*(FL[a]*FLVolume+FR[a]*FRVolume),PByte(K)^);
        Inc(K);
      end;
    end;
  end;
end;

procedure TStereoBufferAudioSource.UnallocateSamples;
begin
  FreeMem(FL);
  FreeMem(FR);
end;

end.

