unit ForwardAudioSource;

interface

uses
  AudioSource,CosineTable,Math,Windows,SysUtils,Classes,Graphics;

const
  FilterSamples=$10000;

type
  TKick=record
    Amplitude,Frequency,StartTime,Length:TFloat;
  end;
  TKicks=array[0..$FFFFFF] of TKick;
  PKicks=^TKicks;

  TForwardAudioSource=class(TMonoBufferAudioSource)
  private
    FKickCount,FKickAllocation:Integer;
    FKicks:PKicks;
    FFilter:array[0..FilterSamples] of TFloat;
    FAbsoluteTime,FRelativeTime:TFloat;
    FLoops,FComplexity:Integer;
    FSpeed,FKickBaseFrequency,FKickAttack,FKickPersistence,FLengthFactor,FHarmonicFactor,FAmplitudeFactor:TFloat;
  protected
    function GetExportLength:TFloat;override;
    procedure Reset;override;

    procedure UnallocateKicks;
    procedure AllocateKicks;

    procedure SampleFilter;
    procedure BuildKicks;

    function GetLoopTime: TFloat;
    function GetPatternTime: TFloat;

    function GetLength:Integer;
    procedure SetLength(const Value:Integer);
    procedure SetComplexity(const Value: Integer);
    procedure SetKickAttack(const Value: TFloat);
    procedure SetKickPersistence(const Value: TFloat);
    procedure SetAmplitudeFactor(const Value: TFloat);
    procedure SetHarmonicFactor(const Value: TFloat);
    procedure SetLengthFactor(const Value: TFloat);
  public
    constructor Create;
    destructor Destroy;override;

    procedure Fill(M:PFloats;Samples,Frequency:Cardinal);override;

    procedure PaintLoop(Canvas:TCanvas;Width,Height:Integer);
    procedure PaintKick(Canvas:TCanvas;Width,Height:Integer);

    property LoopTime:TFloat read GetLoopTime;
    property PatternTime:TFloat read GetPatternTime;

    property Length:Integer read GetLength write SetLength;
    property Complexity:Integer read FComplexity write SetComplexity;
    property Speed:TFloat read FSpeed write FSpeed;

    property KickBaseFrequency:TFloat read FKickBaseFrequency write FKickBaseFrequency;
    property KickAttack:TFloat read FKickAttack write SetKickAttack;
    property KickPersistence:TFloat read FKickPersistence write SetKickPersistence;

    property AmplitudeFactor:TFloat read FAmplitudeFactor write SetAmplitudeFactor;
    property HarmonicalFactor:TFloat read FHarmonicFactor write SetHarmonicFactor;
    property LengthFactor:TFloat read FLengthFactor write SetLengthFactor;
  end;

implementation

function Triangle(X:TFloat):TFloat;
begin
  X:=Frac(X/(2*Pi));
  if X<0.5 then begin
    if X>0.25 then
      Result:=4*(0.5-X)
    else
      Result:=4*X;
  end else begin
    if X>0.75 then
      Result:=4*(X-1)
    else
      Result:=4*(0.5-X);
  end;
end;

function Square(X:TFloat):TFloat;
begin
  X:=Frac(X/(2*Pi));
  if X<0.5 then
    Result:=-1
  else
    Result:=1;
end;

function Accelerate(const t:TFLoat):TFloat;
begin
  Result:=Power(2,t)-1;
end;

function Unaccelerate(const t:TFLoat):TFloat;
begin
  Result:=Ln(t+1)/Ln(2);
end;

{ TForwardAudioSource }

procedure TForwardAudioSource.AllocateKicks;
begin
  if FKickAllocation<FKickCount then begin
    FKickAllocation:=2*FKickCount;
    if Assigned(FKicks) then
      ReallocMem(FKicks,FKickAllocation*SizeOf(TKick))
    else
      GetMem(FKicks,FKickAllocation*SizeOf(TKick));
  end;
end;

procedure TForwardAudioSource.BuildKicks;
var
  a,b,n:Integer;
  TT:array[0..2,0..3] of TFloat;
  U:array[0..2] of TFloat;
const
  TTT:array[0..2,0..3] of TFloat=(
    (0,2,1,3),
    (0,2,1,3),
    (0,1,1,1)
  );

  procedure Rec(s,f,l,t,dt:TFloat;k:Integer;w:Boolean);
  var
    a:Integer;
    u:TFloat;
  begin
    dt:=dt*0.25;
    if k<FComplexity then begin
      for a:=0 to 3 do
        if w or (k<FComplexity-1) or not Odd(a) then
          Rec(TT[0,a]*s,TT[1,a]*f,TT[2,a]*l,t+a*dt,dt,k+1,w);
    end else begin
      if w then begin
        Assert(n<FKickCount);
        with FKicks[n] do begin
          u:=Accelerate(t);
          Amplitude:=u*s+(1-u)*Amplitude;
          Length:=u*l*0.5+(1-u)*Length;
          Length:=Length*(1+u);
          if Odd(n) then begin
            Length:=(1-u)*Length;
            Amplitude:=(1-u)*Amplitude;
          end;
          Frequency:=u*f+(1-u)*Frequency;
        end;
        Inc(n);
      end else begin
        Inc(FKickCount);
        AllocateKicks;
        with FKicks[FKickCount-1] do begin
          Amplitude:=s;
          Length:=l*0.5;
          StartTime:=t;
          Frequency:=f;
        end;
      end;
    end;
  end;

begin
  U[0]:=FAmplitudeFactor;
  U[1]:=FHarmonicFactor;
  U[2]:=FLengthFactor;
  for a:=0 to 2 do
    for b:=0 to 3 do
      TT[a,b]:=Power(U[a],TTT[a,b]);
  FKickCount:=0;
  n:=0;
  for a:=0 to FLoops-1 do begin
    Rec(1,1,0.1/FLoops,a/FLoops,1/FLoops,0,False);
    if Odd(a) then
      Rec(1,1,0.1/FLoops,(a-1)/FLoops,2/FLoops,0,True);
  end;
  Assert(n=FKickCount);
end;

constructor TForwardAudioSource.Create;
begin
  inherited;
  FLoops:=8;
  FComplexity:=3;
  FSpeed:=1;
  FKickBaseFrequency:=100;
  FLengthFactor:=0.9;
  FHarmonicFactor:=1.15;
  FAmplitudeFactor:=0.75;
  FKickAttack:=5;
  FKickPersistence:=0.5;
  SampleFilter;
  BuildKicks;
end;

destructor TForwardAudioSource.Destroy;
begin
  UnallocateKicks;
  inherited;
end;

procedure TForwardAudioSource.Fill(M: PFloats; Samples, Frequency: Cardinal);
var
  a,b:Integer;
  s,t,dt:TFloat;

  function TCos3(const x:TFloat):TFloat;
  begin
    Result:=TCos(x);
    Result:=Result*Result*Result;
  end;

begin
  dt:=1/Frequency;
  for a:=0 to Samples-1 do begin
    s:=0;
    t:=Accelerate(FRelativeTime);
    for b:=0 to FKickCount-1 do
      with FKicks[b] do
        if (StartTime+Length>=1) then begin
          if t>=StartTime then
            s:=s+Amplitude*TCos3(2*Pi*FKickBaseFrequency*Frequency*(FAbsoluteTime+dt*a))*FFilter[Round(FilterSamples*(t-StartTime)/Length)];
          if t+1<=StartTime+Length then
            s:=s+Amplitude*TCos3(2*Pi*FKickBaseFrequency*Frequency*(FAbsoluteTime+dt*a))*FFilter[Round(FilterSamples*(t+1-StartTime)/Length)];
        end else begin
          if (t>=StartTime) and (t<=StartTime+Length) then
            s:=s+Amplitude*TCos3(2*Pi*FKickBaseFrequency*Frequency*(FAbsoluteTime+dt*a))*FFilter[Round(FilterSamples*(t-StartTime)/Length)];
        end;
    FRelativeTime:=Frac(FRelativeTime+FSpeed*0.5*dt/FLoops);
    M[a]:=s;
  end;
  FAbsoluteTime:=FAbsoluteTime+Samples*dt;
end;

function TForwardAudioSource.GetExportLength: TFloat;
begin
  Result:=2*FLoops/FSpeed;
end;

function TForwardAudioSource.GetLength: Integer;
begin
  Result:=FLoops div 2;
end;

function TForwardAudioSource.GetLoopTime: TFloat;
begin
  Result:=FRelativeTime;
end;

function TForwardAudioSource.GetPatternTime: TFloat;
begin
  Result:=Accelerate(FRelativeTime);
  Result:=Frac(Result*FLoops);
end;

procedure TForwardAudioSource.PaintKick(Canvas: TCanvas; Width,
  Height: Integer);
const
  dx1=10;
  dx2=10;
  dy1=10;
  dy2=10;
var
  a:Integer;
  z:TFloat;
begin
  with Canvas do begin
    Brush.Color:=clWhite;
    FillRect(Rect(0,0,Width,Height));
    Pen.Style:=psDot;
    Pen.Color:=clGray;
    MoveTo(0,Height-dy2+1);
    LineTo(Width,Height-dy2+1);
    MoveTo(dx1-1,Height-dy2);
    LineTo(dx1-1,dy1);
    MoveTo(Width-dx2+1,Height-dy2);
    LineTo(Width-dx2+1,dy1);
    Pen.Style:=psSolid;
    Pen.Color:=clRed;
    for a:=dx1 to Width-dx2 do begin
      z:=2*FFilter[Round((a-dx1)*FilterSamples/(Width-dx2-dx1))];
      if a=dx1 then
        MoveTo(a,Height-dy2-Round(z*(Height-dy1-dy2)))
      else
        LineTo(a,Height-dy2-Round(z*(Height-dy1-dy2)));
    end;
  end;
end;

procedure TForwardAudioSource.PaintLoop(Canvas: TCanvas; Width, Height: Integer);
const
  dx1=30;
  dx2=30;
  dy1=10;
  dy2=20;
var
  a,x:Integer;
  t:TFloat;

  function TimeToX(t:TFloat):Integer;
  begin
    Result:=dx1+Round((Width-dx1-dx2)*Unaccelerate(Frac(t)));
  end;

begin
  with Canvas do begin
    FillRect(Rect(0,0,Width,Height));
    Pen.Width:=1;
    Pen.Color:=0;
    Pen.Style:=psSolid;
    MoveTo(0,Height-dy2);
    LineTo(Width,Height-dy2);
    Pen.Style:=psDot;
    Pen.Color:=clGray;
    for a:=0 to FLoops do begin
      t:=a/FLoops;
      x:=TimeToX(t);
      MoveTo(x,Height);
      LineTo(x,0);
    end;
    Pen.Color:=clBlue;
    Pen.Style:=psSolid;
    for a:=0 to FKickCount-1 do
      with FKicks[a] do begin
        t:=StartTime;
        x:=TimeToX(t);
        MoveTo(x,Height-dy2);
        LineTo(x,Height-dy2-Round((Height-dy1-dy2)*Amplitude));
      end;
    Pen.Width:=3;
    Pen.Color:=clRed;
    Pen.Style:=psSolid;
    t:=Accelerate(FRelativeTime);
    x:=TimeToX(t);
    MoveTo(x,Height);
    LineTo(x,0);
  end;
end;

procedure TForwardAudioSource.Reset;
begin
  FAbsoluteTime:=0;
  FRelativeTime:=0;
end;

procedure TForwardAudioSource.SampleFilter;
var
  a:Integer;
  z,s:TFloat;
begin
  z:=FKickAttack/(FKickAttack+FKickPersistence);
  s:=0.5*Power(z,-FKickAttack)*Power(1-z,-FKickPersistence);
  for a:=0 to FilterSamples do begin
    z:=a/FilterSamples;
    FFilter[a]:=s*Power(z,FKickAttack)*Power(1-z,FKickPersistence);
  end;
end;

procedure TForwardAudioSource.SetAmplitudeFactor(const Value: TFloat);
begin
  if not SameValue(FAmplitudeFactor,Value) then begin
    FAmplitudeFactor:=Value;
    BuildKicks;
  end;
end;

procedure TForwardAudioSource.SetComplexity(const Value: Integer);
begin
  if not SameValue(FComplexity,Value) then begin
    FComplexity:=Value;
    BuildKicks;
  end;
end;

procedure TForwardAudioSource.SetHarmonicFactor(const Value: TFloat);
begin
  if not SameValue(FHarmonicFactor,Value) then begin
    FHarmonicFactor:=Value;
    BuildKicks;
  end;
end;

procedure TForwardAudioSource.SetKickAttack(const Value: TFloat);
begin
  if not SameValue(FKickAttack,Value) then begin
    FKickAttack:=Value;
    SampleFilter;
  end;
end;

procedure TForwardAudioSource.SetKickPersistence(const Value: TFloat);
begin
  if not SameValue(FKickPersistence,Value) then begin
    FKickPersistence:=Value;
    SampleFilter;
  end;
end;

procedure TForwardAudioSource.SetLength(const Value: Integer);
begin
  if FLoops<>2*Value then begin
    FLoops:=2*Value;
    BuildKicks;
  end;
end;

procedure TForwardAudioSource.SetLengthFactor(const Value: TFloat);
begin
  if not SameValue(FLengthFactor,Value) then begin
    FLengthFactor:=Value;
    BuildKicks;
  end;
end;

procedure TForwardAudioSource.UnallocateKicks;
begin
  if FKickAllocation>0 then begin
    FreeMem(FKicks);
    FKicks:=nil;
    FKickAllocation:=0;
    FKickCount:=0;
  end;
end;

end.

